// Ripped from https://github.com/Ch0pin/medusa/ and modified to fit Androguard packets

colorLog('[+] LOADING HELPER/ANTIDEBUG/DEBUG.JS',{c: Color.Red});

var antidebug = Java.use('android.os.Debug');

antidebug.isDebuggerConnected.implementation = function(){
    agSysPacket({information: "overwriting is debugger connected"}).send();
    return false;
}
